/*
 *		    _ c	o n r e	a d _ c
 *
 *  This routine is called from	the 'read' call.  It handles
 *  input from the console/screen device.
 *
 *  The	idea of	vectoring the call into	this routine is	that
 *  this makes it possible to replace this routine with	a
 *  full terminal emulator if so desired.    This can be done
 *  in a way that is transparent to user-level programs.
 *
 *
 *  RETURN VALUES:
 *	    0/+ve   This many bytes (from the start of the data
 *		    buffer passed) have	been read
 *		    any	translation.
 *	    -1	    Error occurred.
 *
 *
 *  AMENDMENT HISTORY
 *  ~~~~~~~~~~~~~~~~~
 *  18 Sep 92	DJW	- Original version
 *			  an upgrade to	support	a CURSES library.
 *  11 Jul 93	DJW	- Removed assumption that io_fbyte() sets _oserr.
 *			- Removed assumption that sd_curs() sets _oserr.
 *  8  Apr 94	EAJ	- Corrected treatment of CTRL-D. Just typing CTRL-D
 *			  should return	0 as no. of char's read.
 *  13 Apr 94	EAJ	- vectored the call to io_fbyte. See _readmouse_c.
 *  17 Sep 94	EAJ	- Improved line-editing.
 *  30 Sep 94	EAJ	- Cleaned up a bit. If RLL is defined, the full
 *			  version is produced, which is	ready to be split
 *			  up into 2 parts, one of which	can be placed in
 *			  a RLL. If not, a smaller version will	be produced
 *			  containing the essential editing features only.
 *
 *			  The intention	is that	Dave Walker should include
 *			  the full version in the RLL (when finished),	and
 *			  include the small version in the ordinary library.
 *  21 Dec 94	EAJ	- minor	changes. Still waiting for Dave	...
 *
 *  30 Nov 95	DCG	 Changed the alt in compose char to chr$(0)
 *			  -serious hotkey problems. Also changed  compose
 *			 character to terminate	 on non-hexadecimal
 *			 character returning such if no	valid chars entered
 *
 *
 *  05 Dec 95	DCG	- Adapted for C68 V4.20a changes in UFB	structure
 *
 *  03 Dec 97	DCG	- if error and length >0 return	length instead of EOSERR
 *
 *  18 Dec 98	DCG	- changed from using chr$(0) to	chr$(254) ie shift enter
 *			  for compose character.
 *
 *  04 Mar 02	DCG	- forked off this new file:
 *			   a simpler stripped down version with	no
 *			  support for switching	between	insert and overwite
 *			  no last line recall no ability to send eof at	keyboard
 *			  no compose character.Fairly close to io.edlin	except for tabs
 *			  ie halfway between tyye full and the minimum version.
 *			  also changed issep to	isspace
 *  05 Mar 02	DCG	- Fixed	 bugs geneated by 4 march changes
 *
 *  11 Mar 02	DCG	- added	support	for tab	and shift tab with a fixed
 *			  tab distance of 8 also rejects all characters
 *			  greater than 222.
 *
 *  13 Mar 03	DCG	- made issep simple test for  equality with space
 *			  character. Also optimised tab	calculation a bit
 *			  calculation now tied more tightly with tab distance of 8.
 *
 *
 */
#define	__LIBRARY__
#include <qdos.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
/* #include <ctype.h> */


#define	issep(x) (x==' ')


/*
 *  Used for communication between main	routine	and subroutines.
 *  These next few functions can be placed in a	RLL.
 */
struct info
{
    struct UFB * uptr;
    timeout_t tmout;
    short startx, starty,
	  cursor, curlen, maxlen,
	  winx,	winy,
	  cursx, cursy,
	  pixels;
    unsigned char * buf;
};

static void calpos(struct info *ip)
{
    short tmp;

    tmp	= ip->startx + ip->cursor;
    ip->cursy =	ip->starty + (tmp / ip->winx);
    ip->cursx =	tmp % ip->winx;
}

static void curpos(struct info *ip)
{
    if(	ip->uptr->ufbflg & UFB_NT )
	return;

   (void) sd_pos(ip->uptr->ufbfh, ip->tmout, ip->cursx,	ip->cursy);
}

static void calcurpos(struct info *ip)
{
    calpos(ip);
    curpos(ip);
}


static void scrolldown(short lines, struct info	*ip)
{
    short ofs,len,xpos;

    ip->starty += lines;

    if(	(ip->uptr->ufbflg & UFB_NT) || (lines==0) )
	return;

    /* scroll up */

    (void) sd_scrol(ip->uptr->ufbfh,ip->tmout,ip->pixels*lines);

    /* offset of first character to print */

    ofs	= -( ip->winx *	ip->starty + ip->startx	);

    if(	ofs >= 0 )
    {
	len = ip->winx * lines;
	xpos = 0;
    }
    else
    {
	len = ip->winx * lines + ofs;
	xpos = -ofs;
	ofs = 0;
    }

    (void) sd_pos(ip->uptr->ufbfh, ip->tmout, xpos, 0);

    (void) io_sstrg(ip->uptr->ufbfh, ip->tmout,	ip->buf+ofs, len);
}


static void scrollup(short lines, struct info *ip)
{
    short above,below,maxbelow;

    ip->starty -= lines;

    if(	(ip->uptr->ufbflg & UFB_NT) || (lines==0) )
	return;

    /* scroll up */

    (void) sd_scrol(ip->uptr->ufbfh,ip->tmout,-ip->pixels*lines);

    /* position	cursor at first	character to repaint */

    (void) sd_pos(ip->uptr->ufbfh,ip->tmout, 0,	ip->winy-lines);

    /* number of chars above cursor pos. */

    above = ip->winx * ( ip->winy - lines - ip->starty ) - ip->startx;

    /* maximum number of chars to repaint */

    maxbelow = lines * ip->winx;

    /* actual number of	chars below cursor */

    below = ip->curlen - above;

    if(	below>maxbelow )
	below =	maxbelow;

    if(	below <= 0 )
	return;

    (void) io_sstrg(ip->uptr->ufbfh, ip->tmout,	ip->buf+above, below);
}

static void curstoend(struct info *ip)
{
    short lines;

    ip->cursor=ip->curlen;
    calpos(ip);
    lines=ip->cursy-ip->winy+1;
    if(lines>0)	{
	scrollup(lines,ip);
	calpos(ip);
    }
    curpos(ip);
}


/* This	function contains no static data, and no other funny stuff, so
** it can be placed in a RLL.
*/
static
int  my_default_conread	(struct	UFB * uptr, unsigned char * buf,
			 unsigned int len,
			 int (*myreadkbd)(chanid_t, timeout_t, char *)
			 )
{
    struct info	inf;

    short enq[4],cnt,extra;

    struct REGS	r;

    int	    reply;
    long tmp,tmp2;

    inf.uptr = uptr;
    inf.buf = buf;
    inf.tmout=_GetTimeout(uptr);

    /*
     *	See if we are in EOF state
     */
    if (uptr->ufbst & UFB_EOF) {
	uptr->ufbst &= (char)~UFB_EOF;	  /* Clear EOF flag */
	_oserr = ERR_EF;
	errno =	EOSERR;
	return 0;
    }

    /*
     *	First do any pending newlines, explicitly, for QXL,
     *	since QXL doesn't do it when enabling the cursor
     */
    (void) sd_donl( uptr->ufbfh, inf.tmout);


    inf.maxlen = len;
    inf.cursor = inf.curlen = 0;

    /*
     *	Read window size and cursor position
     */
    r.D0 = 0xb;	r.D3 = -1; r.A0	= (char	*)uptr->ufbfh; r.A1 = (char *)&enq[0];
    (void) qdos3(&r,&r);
    inf.startx = enq[2];
    inf.starty = enq[3];
    inf.winx   = enq[0];
    inf.winy   = enq[1];

    /*
     *	Calculate pixels per line (avoid sd.extop, so network still works)
     */
    r.D0 = 0x10; r.D1 =	r.D2 = 0; (void) qdos3(&r,&r);
    r.D0 = 0x16; (void)	qdos3(&r,&r);
    reply = r.D0;
    r.D0 = 0xa;	r.A1 = (char *)&enq[0];	(void) qdos3(&r,&r);
    if(	reply )
	inf.pixels = enq[1];
    else
	inf.pixels = enq[3];

    /*
     *	Main loop.
     */
    for(;inf.maxlen != inf.curlen;)
    {   unsigned char ch;

	calcurpos(&inf);

	/*  Get	character. myreadkbd  will point to a io_fbyte lookalike function.
	*/
	(void) sd_cure(	uptr->ufbfh, inf.tmout);
	while (1)
       {
	reply =	(*myreadkbd)( uptr->ufbfh, inf.tmout,(char *)&ch);
       if (reply)
       break;
	 if ( (ch==10)||(ch==27)||(ch==9)||(ch==253)|| ( ((ch>31) && (ch<222))	))
	 break;
       }
	(void) sd_curs(	uptr->ufbfh, inf.tmout);

	if( reply )
	{
	    if (inf.curlen > 0)
	    goto DO_EXIT;
	    _oserr=(short)reply;
	    errno = EOSERR;
	    reply=-1;
	    goto KILL_CURSOR;
/*	      break; */
	}



	switch(ch)
	{

	    /*
	     *	Ordinary characters are	inserted into the string here
	     */
	    default:  (void) memmove( buf+inf.cursor+1,	buf+inf.cursor,	inf.curlen-inf.cursor);
			    ++inf.curlen;


			buf[inf.cursor++] = ch;
			cnt = -1; extra=0;

	    PRINT:	/* print from cursor+cnt, with extra trailing char's */
			/* assume display cursor is correct */
			tmp = inf.curlen+extra-inf.cursor-cnt;
			tmp2 = (inf.winy-inf.cursy)*inf.winx-inf.cursx;
			if( tmp2<tmp )
			    tmp=tmp2;
			if( !(uptr->ufbflg & UFB_NT) )
			(void) io_sstrg(uptr->ufbfh, inf.tmout,
				     buf+(inf.cursor+cnt),tmp);
			break;
	    case 9:	/* Tab */
			if(inf.cursor <	inf.curlen )
			{
			 inf.cursor-=((inf.cursor)&7);
			 inf.cursor+=8;
			 if(inf.cursor>inf.curlen)
			 inf.cursor=inf.curlen;
			}
			break;
	    case 253:  /* shift	tab */
		       if(inf.cursor)
		       {
		       if (!(inf.cursor&7))
			 inf.cursor-=8;
			 else
			 inf.cursor-=(inf.cursor&7);
		       }
		       break;

	    case 200:	/* move	right */
			if( inf.cursor < inf.curlen )
			    ++inf.cursor;
			break;


	    case 192:	/* move	left */
			if( inf.cursor )
			     --inf.cursor;
			break;

	    case 195:	/* CTRL	ALT left */
	    case 193:	/* HOME	*/
			cnt=inf.cursor;
			if( inf.cursor < inf.cursx )
			    inf.cursor=0;
			else
			    inf.cursor -= inf.cursx;
			if( ch == 193 )
			    break;
			cnt -= inf.cursor;
			goto LEFT;

	    case 203:	/* CTRL	ALT right */
	    case 201:	/* END */
			cnt=inf.cursor;
			if( inf.cursor+inf.winx-inf.cursx > inf.curlen )
			    inf.cursor=inf.curlen;
			else
			    inf.cursor += inf.winx-inf.cursx;
			if( ch == 201 )
			    break;
			cnt = inf.cursor-cnt;
			inf.cursor -= cnt;
			goto DELETE;

	    case 210:	/* CTRL	UP */
	    case 208:	/* UP */
			cnt=inf.cursor;
			if( inf.cursor-inf.winx<0 )
			    inf.cursor=0;
			else
			    inf.cursor -= inf.winx;
			if( ch == 208 )
			    break;
			cnt -= inf.cursor;
			goto LEFT;

	    case 218:	/* CTRL	DOWN */
	    case 216:	/* DOWN	*/
			cnt=inf.cursor;
			if( inf.cursor+inf.winx>inf.curlen )
			    inf.cursor=inf.curlen;
			else
			    inf.cursor += inf.winx;
			if( ch == 216 )
			    break;
			cnt = inf.cursor-cnt;
			inf.cursor -= cnt;
			goto DELETE;

	    case 27:	/* ESC,	forget all of text */
			inf.cursor=0;
			cnt=inf.curlen;
			goto LEFT;

	    case 212:	/* ALT SHIFT ^,	PGUP */
	    case 213:	inf.cursor=0;
			break;

	    case 220:	/* ALT SHIFT v,	PGDN */
	    case 221:	inf.cursor=inf.curlen;
			break;

	    case 194:	/* delete character left */
			if( inf.cursor ) {
			    --inf.cursor; cnt=1;
			    goto LEFT;
			}
			break;

	    case 202:	/* delete character right */
			if( inf.cursor == inf.curlen )
			    break;
			cnt=1;

	    DELETE:	/* common part of all delete operations	*/
			(void)memmove(buf+inf.cursor, buf+inf.cursor+cnt, inf.curlen-inf.cursor-cnt);
			for(tmp=cnt;--tmp>=0;)
			    buf[--inf.curlen]=' '; /* used to erase tail */

			extra=cnt;
			cnt=0;

			calcurpos(&inf);

			goto PRINT;

	    case 196:	/* move	word left & delete word	left */
	    case 198:	cnt=0;
			while( inf.cursor && issep(buf[inf.cursor-1]) )	{
			    --inf.cursor; ++cnt;
			}
			while( inf.cursor && !issep(buf[inf.cursor-1]))	{
			    --inf.cursor; ++cnt;
			}
	    LEFT:	if( ch == 196 )
			    break;
			while( (inf.startx+inf.cursor)/inf.winx+inf.starty < 0 )
			{   scrolldown(1,&inf); }
			goto DELETE;

	    case 204:	/* move	word right & delete word right */
	    case 206:	cnt=0;
			while( (inf.cursor < inf.curlen) && !issep(buf[inf.cursor]) ) {
			    ++inf.cursor; ++cnt;
			}
			while( (inf.cursor < inf.curlen) && issep(buf[inf.cursor]) ) {
			    ++inf.cursor; ++cnt;
			}
			if( ch == 206 )	{
			    inf.cursor -= cnt;	 /* delete doesn't move cursor */
			    goto DELETE;
			}
			break;

	    case 10:	/* ENTER */

			curstoend(&inf);
			buf[inf.curlen++] = '\n';
			if( !(uptr->ufbflg & UFB_NT) )
			(void) io_sbyte(uptr->ufbfh, inf.tmout,	'\n');
			goto DO_EXIT;

	}

	/*
	 *  Ensure cursor is inside the	window,	by scrolling
	 *  if neccesary.
	 */
	calpos(&inf);
	if( inf.cursy <	0 )
	    scrolldown(-(inf.cursy),&inf);
	else
	    if(	inf.cursy>=inf.winy )
		scrollup(inf.cursy+1-inf.winy,&inf);
    }

    /*
     *	Suppress cursor	when finished
     */


DO_EXIT:
    reply=((int)inf.curlen);
KILL_CURSOR:
    (void) sd_curs( uptr->ufbfh, inf.tmout);
    return (reply);
}


/* This	function transfers pointers to static data to the function
** above, which	may be placed in a RLL.	(hint hint)
*/

/*===================================================================
 *	Default	vector points to this routine
 *------------------------------------------------------------------*/
int	_default_conreadsimple(struct UFB *uptr,
			 unsigned char *buf,
			 unsigned int len)
{

    return my_default_conread(uptr, buf, len,_readkbd);

}
